#ifndef _BSTRING_CPP
#define _BSTRING_CPP

#include <string>
#include <new>

#include "BString.h"

using namespace std;


BString::BString()
{
   bsReserved = bsIncr;

   try {  bsData = new char [bsReserved];  }
   catch( bad_alloc )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
   bsLen = 0;
} // BString::BString

BString::BString( const char* rhs )
{
// cout << "In default constructor; rhs: " << rhs << endl;
   bsLen = cStrLen( rhs );
   bsReserved = ((bsLen+1)/bsIncr+1)*bsIncr;

   try {  bsData = new char [bsReserved];  }
   catch( bad_alloc )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
   for ( int k(0); k < bsLen; ++k )  bsData[k] = rhs[k]; 
} // BString::BString

BString::BString( const BString& rhs )
{
// cout << "In copy constructor; rhs: " << rhs.c_str() << endl;
   bsReserved = rhs.bsReserved;

   try {  bsData = new char [bsReserved];  }
   catch( bad_alloc )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
   for ( int k(0); k < rhs.bsLen; ++k )  bsData[k] = rhs.bsData[k];
   bsLen = rhs.bsLen;
} // BString::BString

inline const char* BString::c_str() const
{  bsData[bsLen] = '\0';  return bsData;  }

char& BString::at( int k ) const
{
   if ( k < 0 || bsLen <= k )
      throw BStringXptn( "at", BStringXptn::outOfRange, k );
   return bsData[k];
} // BString::at

BString& BString::operator+=( const BString& rhs )
{
   if ( bsLen + rhs.bsLen + 1 > bsReserved )
   {
      char* tmp;
      size_t tmpRes( ((bsLen+rhs.bsLen+1)/bsIncr+1)*bsIncr );
      try {  tmp = new char[tmpRes];  }
      catch( bad_alloc )
      {  throw BStringXptn( "operator+=",
                            BStringXptn::allocFailed );  }
      for ( int k(0); k < bsLen; ++k )  tmp[k] = bsData[k];
      delete[] bsData;
      bsData = tmp;
   }
   for ( int j(0), k(bsLen); j < rhs.bsLen; ++j, ++k )
      bsData[k] = rhs.bsData[j];
   bsLen += rhs.bsLen;
   return *this;
} // BString::operator+=
/*
BString& BString::append( const BString& rhs )
{
   if ( bsLen + rhs.bsLen + 1 > bsReserved )
   {
      char* tmp;
      size_t tmpRes( ((bsLen+rhs.bsLen+1)/bsIncr+1)*bsIncr );
      try {  tmp = new char[tmpRes];  }
      catch( bad_alloc )
      {  throw BStringXptn( "append",
                            BStringXptn::allocFailed );  }
      for ( int k(0); k < bsLen; ++k )  tmp[k] = bsData[k];
      delete[] bsData;
      bsData = tmp;
   }
   for ( int j(0), k(bsLen); j < rhs.bsLen; ++j, ++k )
      bsData[k] = rhs.bsData[j];
   bsLen += rhs.bsLen;
   return *this;
} // BString::append
/*
BString& BString::operator=( const BString& rhs )
{
   if ( &rhs != this )  // if ( ! )
   {
   //     
      char* tmp;
      try {  tmp = new char[rhs.bsReserved];  }
      catch( bad_alloc& )
      { throw BStringXptn( "operator=", BStringXptn::allocFailed ); }

      delete[] bsData;
   //     rhs
      bsData = tmp;
      bsReserved = rhs.bsReserved; 
      for ( int k(0); k < rhs.bsLen; ++k )
         bsData[k] = rhs.bsData[k];
      bsLen = rhs.bsLen;
   }
   return *this;
} // BString::operator=
*/
BString& BString::operator=( const BString& rhs )
{
   if ( &rhs != this )
   {
      try {  BString tmp( rhs );
             swap( tmp );  }
      catch( BStringXptn& x )
      {  strcpy( x.funcName, "operator=" );
         throw;  }
   }
   return *this; 
} // BString::operator=
/*
BString& BString::operator=( BString rhs )
{
   if ( &rhs != this )
   {
      try {  swap( rhs );  }
      catch( BStringXptn& x )
      {  strcpy( x.funcName, "operator=" );
         throw;  }
   }
   return *this; 
} // BString::operator=
//*/
void BString::swap( BString& rhs )
{
   size_t sv( bsLen );
   bsLen = rhs.bsLen;  rhs.bsLen = sv; 
   sv = bsReserved;
   bsReserved = rhs.bsReserved;  rhs.bsReserved = sv;

   char* svPtr( bsData );
   bsData = rhs.bsData;  rhs.bsData = svPtr;
} // BString::swap

int BString::compare( const BString& rhs ) const
{
   int fv( stringCmpr(bsData, rhs.bsData,
                      min(bsLen, rhs.bsLen)) );
   if ( fv == 0 )  fv = bsLen - rhs.bsLen;
   return fv;
} // BString::compare
/*
BString& BString::assign( const BString& rhs )
{
   if ( &rhs != this )
   {
      try {  BString tmp( rhs );
             swap( tmp );  }
      catch( BStringXptn& x )
      {  strcpy( x.funcName, "assign" );
         throw;  }
   }
   return *this; 
} // BString::operator=
/*
BString& BString::assign( const BString& rhs )
{
   if ( &rhs != this )  // if ( ! )
   {
   //     
      char* tmp;
      try {  tmp = new char[rhs.bsReserved];  }
      catch( bad_alloc& )
      {  throw BStringXptn( "assign", BStringXptn::allocFailed );  }

      delete[] bsData;
   //     rhs
      bsData = tmp;
      bsReserved = rhs.bsReserved; 
      for ( int k(0); k < rhs.bsLen; ++k )
         bsData[k] = rhs.bsData[k];
      bsLen = rhs.bsLen;
   }
   return *this;
} // BString::assign
*/
//=====================   ======================

size_t BString::cStrLen( const char cs[] )
{
   const char* p( cs );
   while ( *p != '\0' )  ++p;
   return p-cs;
} // BString::cStrLen

int BString::stringCmpr( const char* lhs, const char* rhs, int n )
{
   int k( 0 );
   while ( lhs[k] == rhs[k] && k < n-1 ) ++k;
   return ( static_cast<int>(lhs[k]) - static_cast<int>(rhs[k]) );
} // BString::stringCmpr
/*
ostream& operator<<( ostream& tout, const BString& rhs )
{
   for ( int k(0); k < rhs.bsLen; ++k )  tout << rhs.bsData[k];
   return tout;
} // operator<<( ofstream, BString
*/
ostream& operator<<( ostream& tout, const BString& rhs )
{
   for ( int k(0); k < rhs.length(); ++k )  tout << rhs[k];
   return tout;
} // operator<<( ofstream, BString

bool operator==( const BString& lhs, const BString& rhs )
{
   int fv( BString::stringCmpr(lhs.bsData, rhs.bsData,
                               min(lhs.bsLen, rhs.bsLen)) );
   if ( fv == 0 )  fv = lhs.bsLen - rhs.bsLen;
   return ( fv == 0 );
} // operator==

#endif // _BSTRING_CPP
